#pragma once

#include "2022/cec_problem.hpp"

#include "2022/composition_function_1.hpp"
#include "2022/composition_function_2.hpp"
#include "2022/composition_function_3.hpp"
#include "2022/composition_function_4.hpp"
#include "2022/schaffer_f7.hpp"
#include "2022/hybrid_function_1.hpp"
#include "2022/hybrid_function_2.hpp"
#include "2022/hybrid_function_3.hpp"
#include "2022/levy.hpp"
#include "2022/rastrigin.hpp"
#include "2022/rosenbrock.hpp"
#include "2022/zakharov.hpp"
